@@ -20,17 +20,17 @@ tf.get_logger().propagate = False
 
 from tfx.components import CsvExampleGen
 from typing import Dict, List, Text
-from tfx.components import Evaluator, ExampleValidator, Pusher, ResolverNode, SchemaGen, Trainer, StatisticsGen, Transform
+from tfx.components import Evaluator, ExampleValidator, Pusher, SchemaGen, Trainer, StatisticsGen, Transform
 from tfx.components.base import executor_spec
 from tfx.components.trainer.executor import GenericExecutor
 from tfx.dsl.experimental import latest_blessed_model_resolver
 from tfx.orchestration import metadata, pipeline
 from tfx.orchestration.experimental.interactive.interactive_context import InteractiveContext
-from tfx.proto import pusher_pb2, trainer_pb2
+from tfx.proto import pusher_pb2, trainer_pb2, example_gen_pb2
+from tfx.v1 import proto
 from tfx.types import Channel
 from tfx.types.standard_artifacts import Model
 from tfx.types.standard_artifacts import ModelBlessing
-from tfx.utils.dsl_utils import external_input
 from tfx.orchestration.local.local_dag_runner import LocalDagRunner
 from ml_metadata.proto import metadata_store_pb2
 from tfx.orchestration.portable.mlmd import execution_lib
@@ -118,7 +118,7 @@ class TFAutoData():
 
   def collect_feature_details(self, schema):
     features_list = []
-    features_dict = display_util.get_schema_dataframe(schema).to_dict('index')
+    features_dict = display_util.get_schema_dataframe(schema)[0].to_dict('index')
     features_stats = MessageToDict(self.stats_train)
     self._len_train = features_stats['datasets'][0]['numExamples']
 
@@ -149,9 +149,19 @@ class TFAutoData():
         elif curr_feat == features_dict[f_]['feature'] and features_dict[f_]['Type'] == 'CATEGORICAL':
           features_dict[f_]['categorical_values_count'] = {}
           features_dict[f_]['categorical_values_count_total'] = 0
-          for topvals in feat['stringStats']['topValues']:            
-            features_dict[f_]['categorical_values_count'][topvals.get('value', "NA")] = topvals.get('frequency', 0)
-            features_dict[f_]['categorical_values_count_total'] += topvals.get('frequency', 0)
+          #For each categorical value store counts
+          for categ in features_dict[f_]['categorical_values']:
+            categ_found = 0
+            for topvals in feat['stringStats']['topValues']:
+              if categ == topvals['value']:
+                categ_found = 1
+                features_dict[f_]['categorical_values_count'][topvals.get('value', "NA")] = topvals.get('frequency', 0)
+                features_dict[f_]['categorical_values_count_total'] += topvals.get('frequency', 0)
+                break
+            #If value count for this category not present
+            if categ_found == 0:
+              features_dict[f_]['categorical_values_count'][categ] = 1
+              features_dict[f_]['categorical_values_count_total'] += 1
       
       features_list.append(features_dict[f_])
     self.features_list = features_list
@@ -177,9 +187,13 @@ class TFAutoData():
     for example in dataset.take(1):
       return ([e.numpy().decode('utf-8') for e in example])
 
-  def run_initial(self, _data_path, _tfx_root, _metadata_db_root, tfautils, viz=False):
+  def run_initial(self, _train_data_path, _test_data_path, _tfx_root, _metadata_db_root, tfautils, viz=False):
     """Run all data steps in pipeline and generate visuals"""
-    self.example_gen = CsvExampleGen(input=external_input(_data_path))
+    input = proto.Input(splits=[
+                example_gen_pb2.Input.Split(name='train', pattern=os.path.join(_train_data_path, "*")),
+                example_gen_pb2.Input.Split(name='eval', pattern=os.path.join(_test_data_path, "*"))
+            ])
+    self.example_gen = CsvExampleGen(input_base="/", input_config=input)
 
     self.statistics_gen = StatisticsGen(examples=self.example_gen.outputs['examples'])
 
@@ -230,7 +244,7 @@ class TFAutoData():
     self.features_list = self.collect_feature_details(self.schema)
 
     #Get columns from training file
-    self.file_headers = self.get_columns_from_file_header(_data_path, len(self.features_list))
+    self.file_headers = self.get_columns_from_file_header(_train_data_path, len(self.features_list))
 
     # Visualize results using TFDV
     if viz==True:
@@ -257,21 +271,16 @@ class TFAutoData():
     return self.pipeline
 
 class TextEncoder(tf.keras.Model):
-  def __init__(self, strategy):
+  def __init__(self, strategy, trainable=False):
     self.strategy = strategy
     with self.strategy.scope():
       super(TextEncoder, self).__init__()
-      self.preprocessor = hub.KerasLayer("https://tfhub.dev/google/universal-sentence-encoder-cmlm/multilingual-preprocess/2", trainable=False)
-      self.encoder = hub.KerasLayer("https://tfhub.dev/google/LaBSE/2", trainable=False)
+      self.encoder = hub.KerasLayer("https://tfhub.dev/google/tf2-preview/nnlm-en-dim50/1", trainable=trainable)
 
   def __call__(self, inp):
     with self.strategy.scope():
-      #Preprocess text
-      # encoder_inputs = self.preprocessor(inp)
-
       #Encode text
-      preprocess = self.preprocessor(inp)
-      embedding = self.encoder(preprocess)["default"]
+      embedding = self.encoder(inp)
 
       return embedding
 
@@ -307,7 +316,8 @@ class TFAutoModel():
                 'min':32,
                 'max':32
            },
-           'learning_rate':[0.01]
+           'learning_rate':[0.01],
+           'l1_regularization':[0.00001]
         },
         1:{
             'deep_neurons':{
@@ -322,7 +332,8 @@ class TFAutoModel():
                 'min':32,
                 'max':128
            },
-           'learning_rate':[0.01, 0.005, 0.002, 0.001, 0.0005]
+           'learning_rate':[0.01, 0.005, 0.002, 0.001, 0.0005],
+           'l1_regularization':[0.0001,0.00005,0.00001]
         },
     }
     self.hpt_config[2] = self.hpt_config[1]
@@ -605,18 +616,31 @@ class TFAutoModel():
           feat_numeric.append('')
           
           #Apply normalization
-          feat_numeric[-1] = ( tf.cast(feature_cols['K'][feats['feature']], tf.float32) - feats['mean'] ) / feats['std_dev']
+          if feats['std_dev'] != 0:
+            feat_numeric[-1] = ( tf.cast(feature_cols['K'][feats['feature']], tf.float32) - feats['mean'] ) / feats['std_dev']
+          else:
+            feat_numeric[-1] = tf.cast(feature_cols['K'][feats['feature']], tf.float32)
+          
+          #More feature engineering( Squaring )
+          feat_numeric.append('')
+          feat_numeric[-1] = tf.math.pow(feat_numeric[-2], 2)
 
-      #Handle Text attributes( For model complexity > 2 )
+
+      #Handle Text attributes
       feat_text = []
-      if self._model_complexity >= 2:
+      if self._model_complexity < 2:
+        #Without fine-tuning
         text_emb = TextEncoder(self.strategy)
-        for feats in self._config_json['data_schema']:
-          if feats['Type'] == 'STRING':
-            feat_text.append('')
-            
-            #Apply Text Encoding from TFHub
-            feat_text[-1] = text_emb(tf.squeeze(tf.cast(feature_cols['K'][feats['feature']], tf.string)))
+      else:
+        #With fine-tuning
+        text_emb = TextEncoder(self.strategy, trainable=True)
+
+      for feats in self._config_json['data_schema']:
+        if feats['Type'] == 'STRING':
+          feat_text.append('')
+          
+          #Apply Text Encoding from TFHub
+          feat_text[-1] = text_emb(tf.reshape(tf.cast(feature_cols['K'][feats['feature']], tf.string), [-1]))
 
       ###Create MODEL
       ####Concatenate all features( Numerical input )
@@ -669,8 +693,9 @@ class TFAutoModel():
       prefinal_dense = hp.Int('prefinal_dense', min_value=self.hpt_config[self._model_complexity]['prefinal_dense']['min'], 
                                                 max_value=self.hpt_config[self._model_complexity]['prefinal_dense']['max'],
                               step=32)
+      l1_reg = hp.Choice('l1_regularization', values=self.hpt_config[self._model_complexity]['l1_regularization'], ordered=True)
       x = tf.keras.layers.Dense(prefinal_dense, activation='relu', kernel_initializer="he_uniform",
-                                activity_regularizer=tf.keras.regularizers.l2(0.00001))(x)
+                                activity_regularizer=tf.keras.regularizers.l2(l1_reg))(x)
       x = tf.keras.layers.BatchNormalization()(x)
 
       #Final Layer
@@ -723,18 +748,30 @@ class TFAutoModel():
           feat_numeric.append('')
           
           #apply normalization
-          feat_numeric[-1] = ( tf.cast(feature_cols['K'][feats['feature']], tf.float32) - feats['mean'] ) / feats['std_dev']
+          if feats['std_dev'] != 0:
+            feat_numeric[-1] = ( tf.cast(feature_cols['K'][feats['feature']], tf.float32) - feats['mean'] ) / feats['std_dev']
+          else:
+            feat_numeric[-1] = tf.cast(feature_cols['K'][feats['feature']], tf.float32)
+
+          # #More feature engineering( Squaring )
+          feat_numeric.append('')
+          feat_numeric[-1] = tf.math.pow(feat_numeric[-2], 2)
       
-      #Handle Text attributes( For model complexity > 2 )
+      #Handle Text attributes
       feat_text = []
-      if self._model_complexity >= 2:
+      if self._model_complexity < 2:
+        #Without fine-tuning
         text_emb = TextEncoder(self.strategy)
-        for feats in self._config_json['data_schema']:
-          if feats['Type'] == 'STRING':
-            feat_text.append('')
-            
-            #Apply Text Encoding from TFHub
-            feat_text[-1] = text_emb(tf.squeeze(tf.cast(feature_cols['K'][feats['feature']], tf.string)))
+      else:
+        #With fine-tuning
+        text_emb = TextEncoder(self.strategy, trainable=True)
+
+      for feats in self._config_json['data_schema']:
+        if feats['Type'] == 'STRING':
+          feat_text.append('')
+          
+          #Apply Text Encoding from TFHub
+          feat_text[-1] = text_emb(tf.reshape(tf.cast(feature_cols['K'][feats['feature']], tf.string), [-1]))
 
       ###Create MODEL
       ####Concatenate all features( Numerical input )
@@ -787,8 +824,9 @@ class TFAutoModel():
       prefinal_dense = hp.Int('prefinal_dense', min_value=self.hpt_config[self._model_complexity]['prefinal_dense']['min'], 
                                                 max_value=self.hpt_config[self._model_complexity]['prefinal_dense']['max'],
                               step=32)
+      l1_reg = hp.Choice('l1_regularization', values=self.hpt_config[self._model_complexity]['l1_regularization'], ordered=True)
       x = tf.keras.layers.Dense(prefinal_dense, activation='relu', kernel_initializer="he_uniform",
-                                activity_regularizer=tf.keras.regularizers.l2(0.00001))(x)
+                                activity_regularizer=tf.keras.regularizers.l2(l1_reg))(x)
       x = tf.keras.layers.BatchNormalization()(x)
 
       #Final Layer
@@ -869,6 +907,7 @@ class TFAutoModel():
       tuner = kt.Hyperband(
                       mod_func,
                       objective='val_loss',
+                      overwrite='True',
                       max_epochs=10)
       stop_early = tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=5)
 
@@ -964,9 +1003,10 @@ class TFAutoModel():
     return examples
 
   #Create prediction function to link in WIT
-  def wit_prediction_fn_dyn(self, examples):
+  def wit_prediction_fn_dyn(self, examples, check_mode=False):
     version = "1"
     out = []
+    examples_out = []
 
     #LOCAL: Predict using Keras prediction function
     saved_mod = tf.saved_model.load(self._model_root + "/"  + version)   #Location)
@@ -997,16 +1037,26 @@ class TFAutoModel():
       #Run prediction function on saved model
       # print(keyword_args)
       # break
-      pred = mod_fn(**keyword_args)
+      try:
+        pred = mod_fn(**keyword_args)
 
-      p_ = pred['out'].numpy()
-      out.append(p_[0])
+        p_ = pred['out'].numpy()
+        out.append(p_[0])
+        examples_out.append(ex)
+      except:
+        1 - 1
     
-    return out
+    if check_mode==True:
+      #If we want to find also the valid examples
+      return out, examples_out
+    else:
+      return out
 
   def call_wit(self):
-    #Generate examples for WIT
+    #Generate examples for WIT and also check for valid examples
     examples_wit = self.generate_examples_for_wit()
+    _, examples_wit = self.wit_prediction_fn_dyn(examples_wit, check_mode=True)
+
     if self._model_type == 'REGRESSION':
       wit_type = 'regression'
       config_builder = (WitConfigBuilder(examples_wit, self._config_json['file_headers'])
@@ -1050,6 +1100,7 @@ class TFAutoModel():
     return success_flag
 
   def run_initial(self, label_column, model_type='REGRESSION', model_complexity=1):
+    self.__init__(self._tfx_root, self._train_data_path, self._test_data_path)
     """Run all modeling steps in pipeline and generate results"""
     self._label = label_column
     self._model_type = model_type
@@ -1082,7 +1133,8 @@ class TFAuto():
     self._metadata = os.path.join(self._tfx_root, 'metadata');    # Join ~/tfx/metadata
     self._log_root = os.path.join(self._tfx_root, 'logs');
     self._model_root = os.path.join(self._tfx_root, 'model');
-    self._data_path = train_data_path
+    self._train_data_path = train_data_path
+    self._test_data_path = test_data_path
 
     self._input_fn_module_file = 'inputfn_trainer.py'
     self._constants_module_file = 'constants_trainer.py'
@@ -1121,7 +1173,7 @@ class TFAuto():
     Method to automatically estimate schema of Data
     Viz: (False) Is data visualization required ?
     '''
-    self.pipeline = self.tfadata.run_initial(self._data_path, self._tfx_root, self._metadata_db_root, self.tfautils, viz)
+    self.pipeline = self.tfadata.run_initial(self._train_data_path, self._test_data_path, self._tfx_root, self._metadata_db_root, self.tfautils, viz)
     self.generate_config_json()
 
   def step_model_build(self, label_column, model_type='REGRESSION', model_complexity=1):
@@ -1130,7 +1182,7 @@ class TFAuto():
     Parameters
     label_column: The feature to be used as Label
     model_type: Either of 'REGRESSION', 'CLASSIFICATION'
-    model_complexity: 0 to 2 (0: Model without HPT, 1: Model with HPT, 2: Complexity 1 + Also handle Text features)
+    model_complexity: 0 to 2 (0: Model without HPT, 1: Model with HPT, 2: Complexity 1 + Trainable Text Layer)
     '''
     # #Run Modeling steps
     if self.tfadata._run == True:
